from foresight.environment.environment_info import EnvironmentInfo
from foresight.environment.git.git_helper import GitHelper
from foresight.utils.test_runner_utils import TestRunnerUtils
import os
import logging
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class JenkinsEnvironmentInfoProvider:
    
    ENVIRONMENT = "Jenkins"
    GIT_URL_ENV_VAR_NAME = "GIT_URL"
    GIT_URL_1_ENV_VAR_NAME = "GIT_URL_1"
    GIT_BRANCH_ENV_VAR_NAME = "GIT_BRANCH"
    GIT_COMMIT_ENV_VAR_NAME = "GIT_COMMIT"
    JOB_NAME_ENV_VAR_NAME = "JOB_NAME"
    BUILD_ID_ENV_VAR_NAME = "BUILD_ID"

    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        job_name = os.getenv(cls.JOB_NAME_ENV_VAR_NAME)
        build_id = os.getenv(cls.BUILD_ID_ENV_VAR_NAME)
        test_run_key = TestRunnerUtils.string_concat_by_underscore(job_name, build_id)
        if test_run_key != "":
            return TestRunnerUtils.get_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash, 
                test_run_key)
        else:
            return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)


    @classmethod
    def build_env_info(cls):
        try:
            repo_url = os.getenv(cls.GIT_URL_ENV_VAR_NAME)
            if not repo_url:
                repo_url = os.getenv(cls.GIT_URL_1_ENV_VAR_NAME)
            
            repo_name = GitHelper.extractRepoName(repo_url)
            branch = os.getenv(cls.GIT_BRANCH_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.GIT_COMMIT_ENV_VAR_NAME)
            commit_message = GitHelper.get_commit_message()

            if not branch:
                branch = GitHelper.get_branch()

            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()

            test_run_id = cls.get_test_run_id(repo_url, commit_hash)

            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, repo_name, 
                branch, commit_hash, commit_message)
            print_debug_message_to_console("Jenkins Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("Jenkins Unable to build environment info: {}".format(err))
            LOGGER.error("Jenkins Unable to build environment info: {}".format(err))
            pass
        return None




import abc

ABC = abc.ABCMeta('ABC', (object,), {})


class ContextProvider(ABC):
    @abc.abstractmethod
    def get(self):
        raise Exception("should be implemented")

    @abc.abstractmethod
    def set(self, execution_context):
        raise Exception("should be implemented")

    @abc.abstractmethod
    def clear(self):
        raise Exception("should be implemented")



THUNDRA_APIKEY = 'thundra.apikey'

THUNDRA_DEBUG_ENABLE = 'thundra.agent.debug.enable'

#############################################################################

THUNDRA_DISABLE = 'thundra.agent.disable'
THUNDRA_TRACE_DISABLE = 'thundra.agent.trace.disable'
THUNDRA_METRIC_DISABLE = 'thundra.agent.metric.disable'
THUNDRA_LOG_DISABLE = 'thundra.agent.log.disable'

#############################################################################

THUNDRA_APPLICATION_ID = 'thundra.agent.application.id'
THUNDRA_APPLICATION_INSTANCE_ID = 'thundra.agent.application.instanceid'
THUNDRA_APPLICATION_NAME = 'thundra.agent.application.name'
THUNDRA_APPLICATION_STAGE = 'thundra.agent.application.stage'
THUNDRA_APPLICATION_DOMAIN_NAME = 'thundra.agent.application.domainname'
THUNDRA_APPLICATION_CLASS_NAME = 'thundra.agent.application.classname'
THUNDRA_APPLICATION_VERSION = 'thundra.agent.application.version'
THUNDRA_APPLICATION_TAG_PREFIX = 'thundra.agent.application.tag'
THUNDRA_APPLICATION_REGION = 'thundra.agent.application.region'

#############################################################################

THUNDRA_REPORT_REST_BASEURL = 'thundra.agent.report.rest.baseurl'
THUNDRA_REPORT_CLOUDWATCH_ENABLE = 'thundra.agent.report.cloudwatch.enable'
THUNDRA_REPORT_REST_COMPOSITE_BATCH_SIZE = 'thundra.agent.report.rest.composite.batchsize'
THUNDRA_REPORT_CLOUDWATCH_COMPOSITE_BATCH_SIZE = 'thundra.agent.report.cloudwatch.composite.batchsize'
THUNDRA_REPORT_REST_COMPOSITE_ENABLE = 'thundra.agent.report.rest.composite.enable'
THUNDRA_REPORT_CLOUDWATCH_COMPOSITE_ENABLE = 'thundra.agent.report.cloudwatch.composite.enable'
THUNDRA_REPORT_REST_LOCAL = 'thundra.agent.report.rest.local'

#############################################################################

THUNDRA_LAMBDA_HANDLER = 'thundra.agent.lambda.handler'

THUNDRA_LAMBDA_WARMUP_WARMUPAWARE = 'thundra.agent.lambda.warmup.warmupaware'

THUNDRA_LAMBDA_TIMEOUT_MARGIN = 'thundra.agent.lambda.timeout.margin'

THUNDRA_LAMBDA_ERROR_STACKTRACE_MASK = 'thundra.agent.lambda.error.stacktrace.mask'

THUNDRA_TRACE_REQUEST_SKIP = 'thundra.agent.trace.request.skip'
THUNDRA_TRACE_RESPONSE_SKIP = 'thundra.agent.trace.response.skip'
THUNDRA_LAMBDA_TRACE_KINESIS_REQUEST_ENABLE = 'thundra.agent.lambda.trace.kinesis.request.enable'
THUNDRA_LAMBDA_TRACE_FIREHOSE_REQUEST_ENABLE = 'thundra.agent.lambda.trace.firehose.request.enable'
THUNDRA_LAMBDA_TRACE_CLOUDWATCHLOG_REQUEST_ENABLE = 'thundra.agent.lambda.trace.cloudwatchlog.request.enable'

THUNDRA_LAMBDA_AWS_STEPFUNCTIONS = 'thundra.agent.lambda.aws.stepfunctions'
THUNDRA_LAMBDA_AWS_APPSYNC = 'thundra.agent.lambda.aws.appsync'

#############################################################################

THUNDRA_TRACE_INSTRUMENT_DISABLE = 'thundra.agent.trace.instrument.disable'
THUNDRA_TRACE_INSTRUMENT_TRACEABLECONFIG = 'thundra.agent.trace.instrument.traceableconfig'

#############################################################################

THUNDRA_TRACE_SPAN_LISTENERCONFIG = 'thundra.agent.trace.span.listenerconfig'

#############################################################################

THUNDRA_SAMPLER_TIMEAWARE_TIMEFREQ = 'thundra.agent.sampler.timeaware.timefreq'
THUNDRA_SAMPLER_COUNTAWARE_COUNTFREQ = 'thundra.agent.sampler.countaware.countfreq'

#############################################################################

THUNDRA_TRACE_INTEGRATIONS_AWS_SNS_MESSAGE_MASK = 'thundra.agent.trace.integrations.aws.sns.message.mask'
THUNDRA_TRACE_INTEGRATIONS_AWS_SNS_TRACEINJECTION_DISABLE = 'thundra.agent.trace.integrations.aws.sns.traceinjection.disable'

THUNDRA_TRACE_INTEGRATIONS_AWS_SQS_MESSAGE_MASK = 'thundra.agent.trace.integrations.aws.sqs.message.mask'
THUNDRA_TRACE_INTEGRATIONS_AWS_SQS_TRACEINJECTION_DISABLE = 'thundra.agent.trace.integrations.aws.sqs.traceinjection.disable'

THUNDRA_TRACE_INTEGRATIONS_AWS_LAMBDA_PAYLOAD_MASK = 'thundra.agent.trace.integrations.aws.lambda.payload.mask'
THUNDRA_TRACE_INTEGRATIONS_AWS_LAMBDA_TRACEINJECTION_DISABLE = 'thundra.agent.trace.integrations.aws.lambda.traceinjection.disable'

THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_STATEMENT_MASK = 'thundra.agent.trace.integrations.aws.dynamodb.statement.mask'
THUNDRA_TRACE_INTEGRATIONS_AWS_DYNAMODB_TRACEINJECTION_ENABLE = 'thundra.agent.trace.integrations.aws.dynamodb.traceinjection.enable'

THUNDRA_TRACE_INTEGRATIONS_AWS_ATHENA_STATEMENT_MASK = 'thundra.agent.trace.integrations.aws.athena.statement.mask'

THUNDRA_TRACE_INTEGRATIONS_HTTP_BODY_MASK = 'thundra.agent.trace.integrations.http.body.mask'
THUNDRA_TRACE_INTEGRATIONS_HTTP_URL_DEPTH = 'thundra.agent.trace.integrations.http.url.depth'
THUNDRA_TRACE_INTEGRATIONS_HTTP_TRACEINJECTION_DISABLE = 'thundra.agent.trace.integrations.http.traceinjection.disable'
THUNDRA_TRACE_INTEGRATIONS_HTTP_ERROR_STATUS_CODE_MIN = 'thundra.agent.trace.integrations.http.error.status.code.min'

THUNDRA_TRACE_INTEGRATIONS_REDIS_COMMAND_MASK = 'thundra.agent.trace.integrations.redis.command.mask'
THUNDRA_TRACE_INTEGRATIONS_RDB_STATEMENT_MASK = 'thundra.agent.trace.integrations.rdb.statement.mask'

THUNDRA_TRACE_INTEGRATIONS_ELASTICSEARCH_BODY_MASK = 'thundra.agent.trace.integrations.elasticsearch.body.mask'
THUNDRA_TRACE_INTEGRATIONS_ELASTICSEARCH_PATH_DEPTH = 'thundra.agent.trace.integrations.elasticsearch.path.depth'

THUNDRA_TRACE_INTEGRATIONS_MONGODB_COMMAND_MASK = 'thundra.agent.trace.integrations.mongodb.command.mask'

THUNDRA_TRACE_INTEGRATIONS_EVENTBRIDGE_DETAIL_MASK = 'thundra.agent.trace.integrations.aws.eventbridge.detail.mask'

THUNDRA_TRACE_INTEGRATIONS_AWS_SES_MAIL_MASK = 'thundra.agent.trace.integrations.aws.ses.mail.mask'
THUNDRA_TRACE_INTEGRATIONS_AWS_SES_MAIL_DESTINATION_MASK = 'thundra.agent.trace.integrations.aws.ses.mail.destination.mask'

THUNDRA_TRACE_INTEGRATIONS_HTTP_DISABLE = 'thundra.agent.trace.integrations.http.disable'
THUNDRA_TRACE_INTEGRATIONS_RDB_DISABLE = 'thundra.agent.trace.integrations.rdb.disable'
THUNDRA_TRACE_INTEGRATIONS_AWS_DISABLE = 'thundra.agent.trace.integrations.aws.disable'
THUNDRA_TRACE_INTEGRATIONS_REDIS_DISABLE = 'thundra.agent.trace.integrations.redis.disable'
THUNDRA_TRACE_INTEGRATIONS_ES_DISABLE = 'thundra.agent.trace.integrations.elasticsearch.disable'
THUNDRA_TRACE_INTEGRATIONS_MONGO_DISABLE = 'thundra.agent.trace.integrations.mongodb.disable'
THUNDRA_TRACE_INTEGRATIONS_SQLALCHEMY_DISABLE = 'thundra.agent.trace.integrations.sqlalchemy.disable'
THUNDRA_TRACE_INTEGRATIONS_CHALICE_DISABLE = 'thundra.agent.trace.integrations.chalice.disable'
THUNDRA_TRACE_INTEGRATIONS_DJANGO_DISABLE = 'thundra.agent.trace.integrations.django.disable'
THUNDRA_TRACE_INTEGRATIONS_DJANGO_ORM_DISABLE = 'thundra.agent.trace.integrations.django.orm.disable'
THUNDRA_TRACE_INTEGRATIONS_FLASK_DISABLE = 'thundra.agent.trace.integrations.flask.disable'
THUNDRA_TRACE_INTEGRATIONS_FASTAPI_DISABLE = 'thundra.agent.trace.integrations.fastapi.disable'
THUNDRA_TRACE_INTEGRATIONS_TORNADO_DISABLE = 'thundra.agent.trace.integrations.tornado.disable'

#############################################################################

THUNDRA_LOG_CONSOLE_DISABLE = 'thundra.agent.log.console.disable'
THUNDRA_LOG_LOGLEVEL = 'thundra.agent.log.loglevel'

#############################################################################

THUNDRA_LAMBDA_DEBUGGER_ENABLE = 'thundra.agent.lambda.debugger.enable'
THUNDRA_LAMBDA_DEBUGGER_PORT = 'thundra.agent.lambda.debugger.port'
THUNDRA_LAMBDA_DEBUGGER_LOGS_ENABLE = 'thundra.agent.lambda.debugger.logs.enable'
THUNDRA_LAMBDA_DEBUGGER_WAIT_MAX = 'thundra.agent.lambda.debugger.wait.max'
THUNDRA_LAMBDA_DEBUGGER_IO_WAIT = 'thundra.agent.lambda.debugger.io.wait'
THUNDRA_LAMBDA_DEBUGGER_BROKER_PORT = 'thundra.agent.lambda.debugger.broker.port'
THUNDRA_LAMBDA_DEBUGGER_BROKER_HOST = 'thundra.agent.lambda.debugger.broker.host'
THUNDRA_LAMBDA_DEBUGGER_SESSION_NAME = 'thundra.agent.lambda.debugger.session.name'
THUNDRA_LAMBDA_DEBUGGER_AUTH_TOKEN = 'thundra.agent.lambda.debugger.auth.token'


#############################################################################
THUNDRA_TEST_RUN_ID = "thundra.agent.test.run.id"
THUNDRA_TEST_PROJECT_ID = "thundra.agent.test.project.id"
THUNDRA_TEST_STATUS_REPORT_FREQUENCY = "thundra.agent.test.status.report.freq"
THUNDRA_TEST_LOG_COUNT_MAX = "thundra.agent.test.log.count.max"
THUNDRA_TEST_SPAN_COUNT_MAX = "thundra.agent.test.span.count.max" #TODO it has been not added to README.md
THUNDRA_TEST_ACTIVE = "thundra.agent.test.active"
THUNDRA_TEST_DISABLE = "thundra.agent.test.disable"



import logging
import random
import time

from thundra import constants
from thundra.listeners.thundra_span_listener import ThundraSpanListener

logger = logging.getLogger(__name__)


class LatencyInjectorSpanListener(ThundraSpanListener):
    available_distributions = ["normal", "uniform"]

    def __init__(self, delay=0, variation=0,
                 sigma=0, distribution="uniform",
                 add_info_tags=True):
        self.delay = max(0, delay)
        self.sigma = max(0, sigma)
        self.variation = max(0, variation)
        self.add_info_tags = add_info_tags
        if distribution in self.available_distributions:
            self.distribution = distribution
        else:
            self.distribution = "uniform"

    def on_span_started(self, span):
        self._inject_delay(span)

    def on_span_finished(self, span):
        pass

    def _inject_delay(self, span):
        delay = self._calculate_delay()
        if delay != 0:
            if self.add_info_tags:
                self._add_info_tags(span, delay)
            time.sleep(delay / 1000.0)

    def _calculate_delay(self):
        delay = self.delay
        if self.distribution == "uniform" and self.variation != 0:
            delay = random.randint(self.delay - self.variation, self.delay + self.variation)
        elif self.distribution == "normal" and self.sigma != 0:
            delay = round(random.gauss(self.delay, self.sigma))

        return max(0, delay)

    def _add_info_tags(self, span, injected_delay):
        try:
            info_dict = {
                'type': 'latency_injector_span_listener',
                'injected_delay': injected_delay,
                'delay': self.delay,
                'variation': self.variation,
                'sigma': self.sigma,
                'distribution': self.distribution,
            }
            span.set_tag(constants.THUNDRA_LAMBDA_SPAN_LISTENER_INFO_TAG, info_dict)
        except Exception as e:
            logger.error("error while adding LatencyInjectorSpanListener info tags: %s", e)

    @staticmethod
    def from_config(config):
        kwargs = {}
        delay = config.get('delay')
        variation = config.get('variation')
        sigma = config.get('sigma')
        distribution = config.get('distribution')
        add_info_tags = config.get('addInfoTags')

        if delay is not None:
            try:
                kwargs['delay'] = int(delay)
            except ValueError:
                LatencyInjectorSpanListener._log_value_parse_err(delay, 'delay')
        if variation is not None:
            try:
                kwargs['variation'] = int(variation)
            except ValueError:
                LatencyInjectorSpanListener._log_value_parse_err(variation, 'variation')
        if sigma is not None:
            try:
                kwargs['sigma'] = int(sigma)
            except ValueError:
                LatencyInjectorSpanListener._log_value_parse_err(sigma, 'sigma')
        if distribution is not None:
            try:
                kwargs['distribution'] = str(distribution)
            except ValueError:
                LatencyInjectorSpanListener._log_value_parse_err(distribution, 'distribution')
        if add_info_tags is not None:
            try:
                kwargs['add_info_tags'] = add_info_tags
            except ValueError:
                LatencyInjectorSpanListener._log_value_parse_err(add_info_tags, 'add_info_tags')

        return LatencyInjectorSpanListener(**kwargs)

    @staticmethod
    def _log_value_parse_err(param, param_name):
        logger.error(("couldn't parse %s parameter (%s) of "
                      "LatencyInjectorSpanListener, using the default value"), param_name, param)

    @staticmethod
    def should_raise_exceptions():
        return False



from abc import ABCMeta, abstractmethod

ABC = ABCMeta('ABC', (object,), {'__slots__': ()})


class Propagator(ABC):

    @abstractmethod
    def inject(self, span_context, carrier):
        pass

    @abstractmethod
    def extract(self, carrier):
        pass



from thundra.opentracing.recorder import ThundraRecorder


class ExecutionContext:
    """
    Represents the scope of execution (request, invocation, etc ...)
    and holds scope specific data.
    """

    def __init__(self, **opts):
        self.start_timestamp = opts.get('start_timestamp', 0)
        self.finish_timestamp = opts.get('finish_timestamp', 0)
        self.recorder = opts.get('recorder', ThundraRecorder())
        self.reports = opts.get('reports', [])
        self.transaction_id = opts.get('transaction_id', '')
        self.span_id = opts.get('span_id', '')
        self.trace_id = opts.get('trace_id')
        self.root_span = opts.get('root_span')
        self.scope = opts.get('scope')
        self.invocation_data = opts.get('invocation_data')
        self.user_tags = opts.get('user_tags', {})
        self.tags = opts.get('tags', {})
        self.error = opts.get('error')
        self.user_error = opts.get('user_error')
        self.platform_data = opts.get('platform_data', {})
        self.response = opts.get('response', {})
        self.incoming_trace_links = opts.get('incoming_trace_links', [])
        self.outgoing_trace_links = opts.get('outgoing_trace_links', [])
        self.timeout = opts.get('timeout', False)
        self.logs = opts.get('logs', [])
        self.metrics = opts.get('metrics', {})
        self.capture_log = False
        self.trigger_operation_name = ''
        self.application_resource_name = ''

    def report(self, data):
        """
        Adds data to be reported
        @param data data to be reported
        """
        if isinstance(data, list):
            for report in data:
                self.reports.append(report)
        else:
            self.reports.append(data)

    def get_operation_name(self):
        return ''

    def get_additional_start_tags(self):
        return {}

    def get_additional_finish_tags(self):
        return {}


from opentracing import Format

import thundra.constants as constants
from thundra import utils
from thundra.opentracing.tracer import ThundraTracer
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.base_integration import BaseIntegration


class RequestsIntegration(BaseIntegration):
    CLASS_TYPE = 'http'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        prepared_request = args[0]
        url_dict = utils.parse_http_url(prepared_request.url,
                                        ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_URL_DEPTH))
        return url_dict.get('operation_name')

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        prepared_request = args[0]
        method = prepared_request.method

        url_dict = utils.parse_http_url(prepared_request.url,
                                        ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_URL_DEPTH))
        span = scope.span

        span.domain_name = constants.DomainNames['API']
        span.class_name = constants.ClassNames['HTTP']

        tags = {
            constants.SpanTags['OPERATION_TYPE']: method,
            constants.HttpTags['HTTP_METHOD']: method,
            constants.HttpTags['HTTP_URL']: url_dict.get('url'),
            constants.HttpTags['HTTP_PATH']: url_dict.get('path'),
            constants.HttpTags['HTTP_HOST']: url_dict.get('host'),
            constants.HttpTags['QUERY_PARAMS']: url_dict.get('query'),
            constants.SpanTags['TOPOLOGY_VERTEX']: True,
        }

        span.tags = tags

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_BODY_MASK):
            body = prepared_request.body if prepared_request.body else ""
            scope.span.set_tag(constants.HttpTags["BODY"], body)

        try:
            tracer = ThundraTracer.get_instance()
            tracer.inject(span.context, Format.HTTP_HEADERS, prepared_request.headers)
            prepared_request.headers[constants.TRIGGER_RESOURCE_NAME_KEY] = url_dict.get('operation_name')

            span.set_tag(constants.SpanTags['TRACE_LINKS'], [span.span_id])
        except Exception as e:
            pass

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        super(RequestsIntegration, self).after_call(scope, wrapped, instance, args, kwargs, response, exception)

        if response is not None:
            self.set_response(response, scope.span)
            if response.headers and (response.headers.get("x-amz-apigw-id") or response.headers.get("apigw-requestid")):
                scope.span.class_name = constants.ClassNames['APIGATEWAY']

            if response.headers and response.headers.get(constants.TRIGGER_RESOURCE_NAME_KEY):
                resource_name = response.headers.get(constants.TRIGGER_RESOURCE_NAME_KEY)
                scope.span.operation_name = resource_name

            if (response.status_code and \
                    ConfigProvider.get(
                        config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_ERROR_STATUS_CODE_MIN) <= response.status_code):
                scope.span.set_tag('error.kind', "HttpError")
                scope.span.set_tag('error', True)
                scope.span.set_tag('error.message', response.reason)

    def set_response(self, response, span):
        status_code = response.status_code
        span.set_tag(constants.HttpTags['HTTP_STATUS'], status_code)



from __future__ import absolute_import

import os
import traceback
from importlib import import_module

MODULES = {}
IGNORE_MODULES = ('__init__',)
PYTHON_EXTENSIONS = ('.py', '.pyc')

for module_name in os.listdir(os.path.dirname(__file__)):
    filename, ext = os.path.splitext(module_name)
    if filename in IGNORE_MODULES or ext not in PYTHON_EXTENSIONS:
        continue
    try:
        imported = import_module(".{}".format(filename), __name__)
        MODULES[filename] = imported
    except ImportError:
        traceback.print_exc()


def _import_exists(module_name):
    try:
        import_module(module_name)
        return True
    except ImportError:
        return False


def patch_modules(thundra_instance):
    for module_name, module in MODULES.items():
        if _import_exists(module_name):
            module.patch(thundra_instance)



from thundra.wrappers.fastapi.fastapi_wrapper import FastapiWrapper, logger
from thundra.config import config_names

from thundra.wrappers.fastapi.fastapi_utils import extract_headers

from thundra.config.config_provider import ConfigProvider
from thundra import constants

import traceback

RESPONSE_REDIRECT_STATUS_CODE = 307

def handle_error(exception, exception_handler, execution_context = None):
    error = {
        'type': type(exception).__name__,
        'message': str(exception),
        'traceback': traceback.format_exc()
    }
    exception_handler(error, execution_context)

class ThundraMiddleware(object):
    def __init__(self, app, wrapper):
        self.app = app
        self._wrapper = wrapper

    async def __call__(self, scope, receive, send):
        """ASGI Middleware

        Args:
            scope (Dict): Connection information
            receive (Callable): Get request from client
            send (Callable): Send response to client
            
        """
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        try:
            scope["_thundra_wrapped"] = True
            scope["thundra_execution_context"] = self._wrapper.before_request(scope, None)
            execution_context = scope["thundra_execution_context"]
        except Exception as e:
            logger.error("Error during the before part of Thundra fastapi: {}".format(e))

        def handle_response(message):
            """ Checking response should be redirect or not. If redirected, then pass the response body by
            setting res_redirected field in scope.

            Args:
                message (Response):  
            """
            try:
                if "status" in message and message.get("status") == RESPONSE_REDIRECT_STATUS_CODE:
                    scope["res_redirected"] = True
                if message and message.get("type") == "http.response.start" and message.get("status") != 307:
                    execution_context.response = message
                    execution_context.response["body"] = b""
                    scope["res_redirected"] = False
                elif message and message.get("type") == "http.response.body" and not scope["res_redirected"]:
                    if execution_context.response["body"]:
                        execution_context.response["body"] += message.get("body")
                    else:
                        execution_context.response["body"] = message.get("body")
                    try:
                        if not message.get("more_body") or message.get("more_body") == False:
                            execution_context.response = {
                                "status_code": execution_context.response.get("status"),
                                "headers": extract_headers(execution_context.response),
                                "body": execution_context.response.get("body") if execution_context.response.get("body") else None
                            }
                            self._wrapper.after_request(execution_context)
                    except Exception as e:
                        try:
                            handle_error(e, self._wrapper.error_handler, execution_context)
                        except Exception as exc:
                            logger.error("Error during the after part of Thundra fastapi: {}".format(exc))
            except Exception as e:
                logger.error("Error during getting res body in fast api: {}".format(e))
    

        async def wrapped_send(message):
            handle_response(message)
            await send(message)


        def handle_request(req):
            """Manipulate request for thundra tracer. If request has "more_body" field, then add it
            to current request body in execution context request body.

            Args:
                req (Request): Gathered from asgi receive function
            """
            if req["type"] == "http.request":
                try:
                    if "body" in req:
                        req_body = req.get("body", b"")
                        if execution_context.platform_data["request"]["body"]:
                            execution_context.platform_data["request"]["body"] += req_body
                        else:
                            execution_context.platform_data["request"]["body"] = req_body
                        if not ConfigProvider.get(config_names.THUNDRA_TRACE_REQUEST_SKIP, True):
                            execution_context.root_span.set_tag(constants.HttpTags['BODY'], execution_context.platform_data["request"]["body"])
                except Exception as e:
                    logger.error("Error during getting req body in fast api: {}".format(e))


        async def wrapped_receive():
            """Wrap asgi receive function to get request and add it to thundra tracer

            Raises:
                e: Exception occurs in receive function

            Returns:
                Request: Request gather from asgi receive function
            """
            try:
                req = await receive()
            except Exception as e:
                try:
                    handle_error(e, self._wrapper.error_handler, execution_context)
                except Exception as exc:
                    logger.error("Error during receive request fast api asgi function: {}".format(exc))
                raise e
            handle_request(req)
            return req

        try:
            await self.app(scope, wrapped_receive, wrapped_send)
        except Exception as e:
            try:
                handle_error(e, self._wrapper.error_handler, execution_context)
            except Exception as exc:
                logger.error("Error in the app fastapi: {}".format(exc))
            raise e


from thundra import constants
from thundra.wrappers import wrapper_utils, web_wrapper_utils


def start_trace(plugin_context, execution_context, tracer):
    request = execution_context.platform_data['request']
    request_route_path = str(request.path) if request.path else None

    _request = {
        'method': request.method,
        'host': request.host.split(':')[0],
        'query_params': request.query,
        'body': request.body,
        'headers': request.headers,
        'path': request.path
    }

    web_wrapper_utils.start_trace(execution_context, tracer, 'Tornado', 'API', _request, request_route_path)


def finish_trace(execution_context):
    root_span = execution_context.root_span
    if execution_context.response:
        status_code = get_response_status(execution_context)
        if status_code:
            root_span.set_tag(constants.HttpTags['HTTP_STATUS'], status_code)
        if execution_context.trigger_operation_name and execution_context.response and hasattr(
                execution_context.response, 'headers'):
            execution_context.response.headers[
                constants.TRIGGER_RESOURCE_NAME_KEY] = execution_context.trigger_operation_name
    web_wrapper_utils.finish_trace(execution_context)


def start_invocation(plugin_context, execution_context):
    execution_context.invocation_data = wrapper_utils.create_invocation_data(plugin_context, execution_context)


def finish_invocation(execution_context):
    wrapper_utils.finish_invocation(execution_context)

    # Set response status code
    wrapper_utils.set_response_status(execution_context, get_response_status(execution_context))


def get_response_status(execution_context):
    try:
        status_code = execution_context.response.status_code
    except:
        return None
    return status_code



from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names
from foresight.environment.environment_info_support import EnvironmentSupport
from foresight.model import (TestRunStart , TestRunStatus, TestRunResult, 
                                TestRunFinish, TestRunContext)
from foresight.utils.test_wrapper import TestWrapper
import foresight.utils.generic_utils as utils
import logging, socket, threading

LOGGER = logging.getLogger(__name__)

class _TestRunScope:
    
    def __init__(self, id=None, task_id = None, start_timestamp = None, context = None):

        self.id = id
        self.task_id = task_id
        self.start_timestamp = start_timestamp
        self.context = context

class _StatusReporter:

    def __init__(self):
        self.delay = ConfigProvider.get(config_names.THUNDRA_TEST_STATUS_REPORT_FREQUENCY)
        self.t = None


    def start(self, start_immediately=False):
        if not self.t:
            if start_immediately:
                self.t = threading.Timer(0, self.report_status)
            else:
                self.t = threading.Timer(self.delay, self.report_status)
            self.t.daemon = True
            self.t.start()


    def stop(self):
        if self.t:
            try:
                if self.t.is_alive():
                    self.t.cancel()
            except Exception as e:
                utils.print_debug_message_to_console("Error occured while canceling Status Reporter")
                pass
            finally:
                self.t = None


    def report_status(self):
        try:
            test_wrapper = TestWrapper.get_instance()
            status_time = utils.current_milli_time()
            test_run_status = TestRunStatus(
                id = TestRunnerSupport.test_run_scope.id,
                project_id = TestRunnerSupport.get_project_id_from_config(),
                task_id = TestRunnerSupport.test_run_scope.task_id,
                start_timestamp= TestRunnerSupport.test_run_scope.start_timestamp,
                status_timestamp= status_time,
                total_count= TestRunnerSupport.test_run_scope.context.total_count.value,
                successful_count= TestRunnerSupport.test_run_scope.context.successful_count.value,
                failed_count = TestRunnerSupport.test_run_scope.context.failed_count.value,
                ignored_count= TestRunnerSupport.test_run_scope.context.ignored_count.value,
                aborted_count= TestRunnerSupport.test_run_scope.context.aborted_count.value,
                host_name = TestRunnerSupport.HOST_NAME,
                environment_info= EnvironmentSupport.environment_info,
                # TODO tags
            )
            test_wrapper.send_test_run_data(test_run_status)
        except Exception as err:
            LOGGER.error("Couldn't send test run status: ".format(err))
        finally:
            self.stop()
            self.start()


class TestRunnerSupport:

    HOST_NAME = socket.gethostname()

    # TODO 
    '''
        For now, this design not support the parallell execution. 
        Test suite execution context and test case execution context may be 
        stored in pytest item object.
    '''
    test_suite_execution_context = None
    test_suite_application_info = None
    test_run_scope = None
    status_reporter = None
    
    @staticmethod
    def get_project_id_from_config():
        return ConfigProvider.get(config_names.THUNDRA_TEST_PROJECT_ID)

    @staticmethod
    def get_project_max_span_count():
        return ConfigProvider.get(config_names.THUNDRA_TEST_SPAN_COUNT_MAX)


    @classmethod
    def set_test_suite_execution_context(cls, execution_context):
        cls.test_suite_execution_context = execution_context


    @classmethod
    def set_test_suite_application_info(cls, application_info):
        cls.test_suite_application_info = application_info


    @classmethod
    def clear_test_run(cls):
        cls.test_run_scope = None


    @classmethod
    def clear_state(cls):
        cls.test_suite_execution_context = None
        cls.test_suite_application_info = None


    @staticmethod
    def do_get_test_run_id():
        test_run_id = None
        try:
            environment_info = EnvironmentSupport.environment_info
            if environment_info != None:
                test_run_id = environment_info.test_run_id
            if test_run_id == None:
                test_run_id = utils.create_uuid4()
        except Exception as err:
            LOGGER.error("Couldn't get test run id: {}".format(err))
            pass
        return test_run_id


    @classmethod
    def start_test_run(cls):
        try:
            test_wrapper = TestWrapper.get_instance()
            context = TestRunContext()
            id = cls.do_get_test_run_id()
            task_id = utils.create_uuid4()
            current_time = utils.current_milli_time()
            cls.test_run_scope = _TestRunScope(id, task_id, current_time, context)
            environment = EnvironmentSupport.environment_info if EnvironmentSupport.environment_info != None else None
            test_run_start = TestRunStart(
                cls.test_run_scope.id,
                cls.get_project_id_from_config(),
                cls.test_run_scope.task_id,
                cls.test_run_scope.start_timestamp,
                cls.HOST_NAME,
                environment
            )
            test_wrapper.send_test_run_data(test_run_start)
            if cls.status_reporter:
                cls.status_reporter.stop()
            else:
                cls.status_reporter = _StatusReporter()
            cls.status_reporter.start(start_immediately=True)
        except Exception as err:
            LOGGER.error("Couldn't start test run properly: {}".format(err))
            pass
    

    @classmethod
    def finish_current_test_run(cls):
        try:
            test_run_context = cls.get_test_run_context()
            if test_run_context:
                test_run_result = TestRunResult(
                    total_count = test_run_context.total_count.value,
                    successful_count = test_run_context.successful_count.value,
                    failed_count = test_run_context.failed_count.value,
                    ignored_count = test_run_context.ignored_count.value,
                    aborted_count = test_run_context.aborted_count.value
                ) 
                cls.finish_test_run(test_run_result)
        except Exception as err:
            LOGGER.error("Couldn't finish test run properly: {}".format(err))
            pass


    @classmethod
    def finish_test_run(cls, test_run_result):
        try:
            test_wrapper = TestWrapper.get_instance()
            finish_time = utils.current_milli_time()
            if cls.test_run_scope:
                if cls.status_reporter:
                    try:
                        cls.status_reporter.stop()
                    finally:
                        cls.status_reporter = None
                test_run_finish = TestRunFinish(
                    id = cls.test_run_scope.id,
                    project_id = cls.get_project_id_from_config(),
                    task_id = cls.test_run_scope.task_id,
                    start_timestamp= cls.test_run_scope.start_timestamp,
                    finish_timestamp= finish_time,
                    duration = finish_time - cls.test_run_scope.start_timestamp,
                    total_count= test_run_result.total_count,
                    successful_count= test_run_result.successful_count,
                    failed_count = test_run_result.failed_count,
                    ignored_count= test_run_result.ignored_count,
                    aborted_count= test_run_result.aborted_count,
                    host_name = cls.HOST_NAME,
                    environment_info= EnvironmentSupport.environment_info,
                    # TODO tags
                )
                test_wrapper.send_test_run_data(test_run_finish)
        except Exception as err:
            LOGGER.error("Thundra foresight finist test run error: {}".format(err))
            pass
        finally:
            cls.test_run_scope = None


    @classmethod
    def get_test_run_context(cls):
        if cls.test_run_scope:
            return cls.test_run_scope.context
        return None


    @classmethod 
    def get_test_run_id(cls):
        if cls.test_run_scope:
            return cls.test_run_scope.id
        return None


    @classmethod
    def get_test_run_task_id(cls):
        if cls.test_run_scope.task_id:
            return cls.test_run_scope.task_id
        return None


import base64
import json
import logging
import re
from gzip import GzipFile

try:
    from BytesIO import BytesIO
except ImportError:
    from io import BytesIO
from thundra.listeners import ThundraSpanListener

from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names

logger = logging.getLogger(__name__)

_active_span_listeners = []
_sampler = None

SPAN_LISTENERS = {
    sl_class.__name__: sl_class
    for sl_class in ThundraSpanListener.__subclasses__()
}

# To match span listener class type and it's arguments as two groups
r1 = r'(\w+)\[(.*)\]'
# To match each argument key value pair in arguments list
r2 = r'[\w.]+=[\w.\!\?\-\"\'\:\(\) ]*'
# Compile reg exs
p1 = re.compile(r1)
p2 = re.compile(r2)


def get_span_listeners():
    return _active_span_listeners


def register_span_listener(listener):
    _active_span_listeners.append(listener)


def clear_span_listeners():
    global _active_span_listeners
    _active_span_listeners = []


def _get_sl_class(sl_class_name):
    sl_class = None
    if sl_class_name is not None:
        try:
            sl_class = SPAN_LISTENERS[sl_class_name]
        except KeyError:
            logger.error('given span listener class %s is not found', sl_class_name)
    return sl_class


def _parse_span_listeners():
    # Clear before parsing to prevent appending duplicate span listeners
    clear_span_listeners()
    # Add span listeners configured using environment variables
    for env_k in ConfigProvider.configs:
        env_v = ConfigProvider.get(env_k)
        if env_k.startswith(config_names.THUNDRA_TRACE_SPAN_LISTENERCONFIG):
            try:
                # Not in JSON format. Should be zipped + encoded. Decode + unzip it
                if not env_v.startswith('{'):
                    compressed_env = base64.b64decode(env_v)
                    env_v = str(GzipFile(fileobj=BytesIO(compressed_env)).read(), 'utf-8')
                span_listener_config_json = json.loads(env_v)

                listener_type = span_listener_config_json.get("type")
                if not listener_type:
                    raise Exception("type property is mandatory in " + env_k + " configuration")

                listener_config = span_listener_config_json.get("config")
                if not listener_config:
                    raise Exception("config property is mandatory in " + env_k + " configuration")

                span_listener_class = _get_sl_class(listener_type)

                if span_listener_class is not None:
                    span_listener = span_listener_class.from_config(listener_config)
                    register_span_listener(span_listener)

            except Exception as e:
                logger.error(("couldn't parse environment variable %s "
                              "to create a span listener"), env_k)


# Parse span listeners from environment variables
_parse_span_listeners()


def get_sampler():
    return _sampler


def set_sampler(sampler):
    global _sampler
    _sampler = sampler



import logging
import uuid

from opentracing import Format

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.plugins.invocation import invocation_support, invocation_trace_support
from thundra.utils import get_normalized_path
import thundra.wrappers.cp_wrapper_utils as cp_wrapper_utils

Logger = logging.getLogger(__name__)

def start_trace(execution_context, tracer, class_name, domain_name, request, request_route_path=None):
    propagated_span_context = tracer.extract(Format.HTTP_HEADERS, request.get('headers'))
    trace_id = str(uuid.uuid4())
    incoming_span_id = None
    if propagated_span_context:
        trace_id = propagated_span_context.trace_id
        incoming_span_id = propagated_span_context.span_id

    # Start root span
    url_path_depth = ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_URL_DEPTH)
    normalized_path = get_normalized_path(request.get('path'), url_path_depth)
    operation_name = request_route_path or normalized_path
    scope = tracer.start_active_span(operation_name=operation_name,
                                     child_of=propagated_span_context,
                                     start_time=execution_context.start_timestamp,
                                     finish_on_close=False,
                                     trace_id=trace_id,
                                     transaction_id=execution_context.transaction_id,
                                     execution_context=execution_context)
    root_span = scope.span

    # Set root span class and domain names
    root_span.class_name = class_name
    root_span.domain_name = domain_name

    # Add root span tags
    execution_context.span_id = root_span.context.span_id
    root_span.on_started()
    root_span.set_tag(constants.HttpTags['HTTP_METHOD'], request.get('method'))
    root_span.set_tag(constants.HttpTags['HTTP_HOST'], request.get('host', ''))
    root_span.set_tag(constants.HttpTags['QUERY_PARAMS'], request.get('query_params'))
    root_span.set_tag(constants.HttpTags['HTTP_PATH'], request.get('path'))
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_REQUEST_SKIP, True):
        root_span.set_tag(constants.HttpTags['BODY'], request.get('body'))
    execution_context.root_span = root_span
    execution_context.scope = scope
    execution_context.trace_id = trace_id

    if request_route_path:
        trigger_operation_name = request.get('host', '') + request_route_path
    else:
        trigger_operation_name = request.get('headers').get(constants.TRIGGER_RESOURCE_NAME_KEY) or \
                             request.get('host', '') + normalized_path
    execution_context.application_resource_name = request_route_path or normalized_path
    invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES'], [trigger_operation_name])
    execution_context.trigger_operation_name = trigger_operation_name

    invocation_support.set_agent_tag(constants.HttpTags['HTTP_METHOD'], request.get('method'))
    invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME'], 'API')
    invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME'], 'HTTP')

    if incoming_span_id:
        invocation_trace_support.add_incoming_trace_link(incoming_span_id)


def update_application_info(application_info_provider, application_info, app_class_name):
    application_info_provider.update({
        'applicationName': application_info.get('applicationName', 'thundra-app'),
        'applicationClassName': app_class_name,
        'applicationDomainName': 'API',
        'applicationInstanceId': application_info.get('applicationInstanceId',
                                                      str(uuid.uuid4())),
        'applicationId': 'python:{}:{}:{}'.format(app_class_name,
                                                  application_info.get('applicationRegion', ''),
                                                  application_info.get('applicationName',
                                                                       'thundra-app'))
    })


def process_request_route(execution_context, request_route_path, request_host):
    if request_route_path and execution_context and execution_context.scope:
        trigger_operation_name = request_host + request_route_path
        execution_context.scope.span.operation_name = request_route_path
        execution_context.trigger_operation_name = trigger_operation_name
        execution_context.application_resource_name = request_route_path
        invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES'], [trigger_operation_name])


def finish_trace(execution_context):
    root_span = execution_context.root_span
    scope = execution_context.scope
    try:
        root_span.finish(f_time=execution_context.finish_timestamp)
        cp_wrapper_utils.on_finish(execution_context, execution_context.platform_data["request"],
                                   execution_context.response, root_span)
    except Exception:
        # TODO: handle root span finish errors
        pass
    finally:
        try:
            scope.close()
        except Exception as e:
            Logger.debug("Error occured while closing scope: {}".format(e))
            pass



import logging
from thundra.listeners import ThundraSpanListener
from thundra.listeners.thundra_span_filterer import SimpleSpanFilter, SpanFilterer
from thundra.listeners.thundra_span_filterer import StandardSpanFilterer
from thundra.listeners.composite_span_filter import CompositeSpanFilter

logger = logging.getLogger(__name__)

class FilteringSpanListener(ThundraSpanListener):
    
    def __init__(self, listener=None, filterer=None):
        self.listener = listener
        self.filterer = filterer

    def on_span_started(self, span):
        if self.listener is None:
            return
        if self.filterer is None or self.filterer.accept(span):
            self.listener.on_span_started(span)
    
    def on_span_finished(self, span):
        if self.listener is None:
            return
        if self.filterer is None or self.filterer.accept(span):
            self.listener.on_span_finished(span)
    
    @staticmethod
    def should_raise_exceptions():
        return True

    @staticmethod
    def from_config(config):
        listener = FilteringSpanListener._get_span_listener_from_config(config)
        filterer = FilteringSpanListener._get_span_filterer_from_config(config)
        return FilteringSpanListener(listener=listener, filterer=filterer)

    @staticmethod
    def _get_span_filters_from_config(config):
        span_filters = []
        is_all = config.get('all', False)
        if not config.get('filters'):
            return StandardSpanFilterer(span_filters=[], all_mandatory=is_all)
        for filter_config in config.get('filters'):
            composite = filter_config.get('composite')
            if composite:
                composite_filter = CompositeSpanFilter(is_all=filter_config.get('all', False))
                filters = FilteringSpanListener._get_span_filters_from_config(filter_config)
                composite_filter.set_filters(filters)
                span_filters.append(composite_filter)
            else:
                span_filters.append(SimpleSpanFilter(
                    filter_config.get('domainName'),
                    filter_config.get('className'),
                    filter_config.get('operationName'),
                    filter_config.get('tags'),
                    filter_config.get('reverse')
                ))
        
        return span_filters

    @staticmethod
    def _get_span_filterer_from_config(config):
        SPAN_FILTERERS= {
            sl_class.__name__: sl_class 
            for sl_class in SpanFilterer.__subclasses__()
        }
        is_all = config.get('all', False)

        filterer_class = None
        if 'filterer' in config:
            try:
                filterer_class_name = config.get("filterer")
                filterer_class = SPAN_FILTERERS[filterer_class_name]
            except KeyError:
                filterer_class = None
                logger.error('given span listener class %s is not found', filterer_class)
                return None
        else:
            filterer_class = StandardSpanFilterer
        span_filters = FilteringSpanListener._get_span_filters_from_config(config)
        return filterer_class(span_filters=span_filters, all_mandatory=is_all)

    @staticmethod
    def _get_span_listener_from_config(config):
        SPAN_LISTENERS = {
            sl_class.__name__: sl_class 
            for sl_class in ThundraSpanListener.__subclasses__()
        }
        listener_config = None
        if 'listener' not in config:
            listener_class = None
        else:
            listener = config.get('listener')
            listener_type = listener.get('type')
            listener_config = listener.get('config')
            try:
                listener_class = SPAN_LISTENERS[listener_type]
            except KeyError:
                listener_class = None
                logger.error('given span listener class %s is not found', listener_type)

        listener = None
        if listener_class is not None:
            if listener_config is not None:
                listener = listener_class.from_config(listener_config)
            else:
                return listener_class()
        return listener



from thundra.wrappers.flask.flask_wrapper import FlaskWrapper, logger
from thundra.utils import Singleton

class ThundraMiddleware(Singleton):
    def __init__(self):
        self._wrapper = FlaskWrapper()

    def set_app(self, app):
        app.before_request(self.before_request)
        app.after_request(self.after_request)
        app.teardown_request(self.teardown_request)

    def before_request(self):
        try:
            from flask import request
            setattr(request, '_thundra_wrapped', True)
            self._wrapper.before_request(request)
        except Exception as e:
            logger.error("Error during the before part of Thundra: {}".format(e))

    def after_request(self, response):
        try:
            self._wrapper.after_request(response)
        except Exception as e:
            logger.error("Error setting response to context for Thundra: {}".format(e))
        return response

    def teardown_request(self, exception):
        try:
            self._wrapper.teardown_request(exception)
        except Exception as e:
            logger.error("Error during the request teardown of Thundra: {}".format(e))



from thundra import constants
from thundra.integrations.rdb_base import RdbBaseIntegration
from thundra.integrations.base_integration import BaseIntegration
from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names

try:
    from psycopg2.extensions import parse_dsn
except ImportError:
    def parse_dsn(dsn):
        return dict(
            attribute.split("=") for attribute in dsn.split()
            if "=" in attribute
        )


class PostgreIntegration(BaseIntegration, RdbBaseIntegration):
    CLASS_TYPE = 'postgresql'
    
    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        return parse_dsn(instance.dsn).get('dbname', '')

    def before_call(self, scope, cursor, connection, _args, _kwargs, response, exception):
        span = scope.span
        span.domain_name = constants.DomainNames['DB']
        span.class_name = constants.ClassNames['POSTGRESQL']

        dsn = parse_dsn(connection.dsn)
        
        query = ''
        operation = ''
        try:
            query = _args[0]
            if len(query) > 0:
                operation = query.split()[0].strip("\"").lower()
        except Exception:
            pass

        tags = {
            constants.SpanTags['OPERATION_TYPE']: PostgreIntegration._OPERATION_TO_TYPE.get(operation, ''),
            constants.SpanTags['DB_INSTANCE']: dsn.get('dbname', ''),
            constants.SpanTags['DB_HOST']: dsn.get('host', ''),
            constants.SpanTags['DB_TYPE']: "postgresql",
            constants.SpanTags['DB_STATEMENT_TYPE']: operation.upper(),
            constants.SpanTags['TRIGGER_CLASS_NAME']: "API",
            constants.SpanTags['TOPOLOGY_VERTEX']: True
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_STATEMENT_MASK):
            tags[constants.DBTags['DB_STATEMENT']] = query

        span.tags = tags




import abc
import logging
import time
import traceback

from thundra.opentracing.tracer import ThundraTracer

logger = logging.getLogger(__name__)

ABC = abc.ABCMeta('ABC', (object,), {})


class BaseIntegration(ABC):
    def run_and_trace(self, wrapped, instance, args, kwargs):
        tracer = ThundraTracer.get_instance()
        if not tracer.get_active_span():
            return wrapped(*args, **kwargs)

        response = None
        exception = None

        scope = tracer.start_active_span(operation_name=self.get_operation_name(wrapped, instance, args, kwargs),
                                         finish_on_close=False)
        # Inject before span tags
        try:
            self.before_call(scope, wrapped, instance, args, kwargs, response, exception)
        except Exception as instrumentation_exception:
            e = {
                'type': str(type(instrumentation_exception)),
                'message': str(instrumentation_exception),
                'traceback': traceback.format_exc(),
                'time': time.time()
            }
            scope.span.set_tag('instrumentation_error', e)

        try:
            # Inform span that initialization completed
            scope.span.on_started()
            # Call original
            response = self.actual_call(wrapped, args, kwargs)
        except Exception as e:
            exception = e

        # Inject after span tags 
        try:
            self.after_call(scope, wrapped, instance, args, kwargs, response, exception)
        except Exception as instrumentation_exception:
            e = {
                'type': str(type(instrumentation_exception)),
                'message': str(instrumentation_exception),
                'traceback': traceback.format_exc(),
                'time': time.time()
            }
            scope.span.set_tag('instrumentation_error', e)

        try:
            scope.span.finish()
        except Exception as e:
            if exception is None:
                exception = e
            else:
                logger.error(e)

        scope.close()

        if exception is not None:
            scope.span.set_error_to_tag(exception)
            raise exception

        return response

    def actual_call(self, wrapped, args, kwargs):
        return wrapped(*args, **kwargs)

    def set_exception(self, exception, traceback_data, span):
        span.set_tag('error.stack', traceback_data)
        span.set_error_to_tag(exception)

    @abc.abstractmethod
    def get_operation_name(self, wrapped, instance, args, kwargs):
        raise Exception("should be implemented")

    @abc.abstractmethod
    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        raise Exception("should be implemented")

    def after_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        if exception is not None:
            self.set_exception(exception, traceback.format_exc(), scope.span)



from thundra.utils import Singleton
import foresight.utils.generic_utils as utils
import logging, threading

logger = logging.getLogger(__name__)

class ThreadExecutorTerminator(threading.Thread):
    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self)
        self.event = threading.Event()
        
    def run(self):
        terminator = Terminator()
        terminator._wait()

class Terminator(Singleton):
    def __init__(self):
        self.tasks = []


    def register_task(self, task): # task => BaseWrapper 
        self.tasks.append(task)


    def _wait(self):
        for task in self.tasks:
            try:
                task.thread_pool_executor.shutdown(wait=True)
            except Exception as e:
                logger.error(f"Task wait error in Terminator".format(e))


    def wait(self, timeout=30):
        terminator_thread = ThreadExecutorTerminator()
        terminator_thread.start()
        terminator_thread.join(timeout)
        if terminator_thread.is_alive():
            logger.debug("Thread is killed by event!")
            terminator_thread.event.set()
        else:
            logger.debug("Thread has already finished!")


import abc

ABC = abc.ABCMeta('ABC', (object,), {})


class ThundraSpanListener(ABC):
    @abc.abstractmethod
    def on_span_started(self, operation_name):
        raise Exception("should be implemented")

    @abc.abstractmethod
    def on_span_finished(self, span):
        raise Exception("should be implemented")

    @staticmethod
    @abc.abstractmethod
    def from_config(config):
        raise Exception("should be implemented")

    @staticmethod
    @abc.abstractmethod
    def should_raise_exceptions():
        raise Exception("should be implemented")
